# -*- coding: iso-8859-1 -*-
"""
This module implements TomTom's chkfile format
"""

__version__ = '$Id$'

import os, struct, types
from cStringIO import StringIO, InputType

import tompress

try:
	from tomcrypt import CipherFile
except:
	class CipherFile:
		def __init__(self, handle, passphrase=None):
			self.__filehandle = handle
			if passphrase is not None:
				raise IOError('Sorry, encryption is not available.')
		def __getattr__(self, method):
			return getattr(self.__filehandle, method)
		def crypt(self, passphrase=None):
			if passphrase is not None:
				raise IOError('Sorry, encryption is not available.')
			return False
		def plain(self):
			return False


#
# TomTom fileformat specs
#
CHK_COUNT = '>L'    # chk-file element count
CHK_INDEX = '>L'    # chk-file element position
ITEM_HEAD = '>BBH'  # element header (typ, compression, blocks)
FILE_HEAD = '>LLHH' # file header tags


#
# Struct helper functions
#
def sread(spec, handle):
	"""
	Reads and decodes struct spec'd bytes from filehandle.

	@param spec: struct specification
	@type  spec: String
	@param handle: filehandle to read from
	@type  handle: Filehandle

	@return: Data tuple
	@rtype : Tuple ( Various )
	"""
	return struct.unpack(spec, handle.read(slen(spec)))

def swrite(spec, handle, *data):
	"""
	Encodes and writes struct spec'd bytes to filehandle.

	@param spec: struct specification
	@type  spec: String
	@param handle: filehandle to write to
	@type  handle: Filehandle

	@return: None
	@rtype : None
	"""
	handle.write(struct.pack(spec, *data))

LEN_CACHE = {}
def slen(spec):
	"""
	Calculates and caches the length of the given struct spec.

	@param spec: struct specification
	@type  spec: String

	@return: length of struct in bytes
	@rtype : Int
	"""
	try:
		return LEN_CACHE[spec]
	except KeyError:
		size = struct.calcsize(spec)
		LEN_CACHE.setdefault(spec, size)
		return size


#
# Main Functions
#
def open(filename, passphrase=None):
	"""
	Open wrapper. Creates ChkFile instance for easy file-like usage.

	@param filename: filename of chkfile to open
	@type  filename: String
	@param passphrase: passphrase to use
	@type  passphrase: String or None

	@return: ChkFile instance
	@rtype : ChkFile
	"""
	return ChkFile(filename, passphrase)


#
# Main CHK-File class
#
class ChkFile:

	"""
	ChkFile class handles TomTom CHK file containers
	"""

	__filetags     = [0xffffffff, 0xffffffff, 0xffff, 0xffff]
	__itemtags     = [0x01]
	__maxpos       = 0xffffffff
	__maxcount     = 0x7fffffff

	__comp_order   = [0x04, 0x00]
	__compressions = {
		0x00: (tompress.uncompress_plain,  tompress.compress_plain ),
		0x04: (tompress.uncompress_tomtom, tompress.compress_tomtom),
	}

	def __init__(self, filename, passphrase=None):

		self.__filename   = filename
		self.__passphrase = passphrase
		self.__changed    = False
		self.__positions  = []
		self.__elements   = []
		self.__closed     = False
		self.__handle     = None

		try:
			self.__handle = CipherFile( \
			                  file(filename, 'rb+'), \
			                  self.__passphrase \
			                )
		except IOError as reason:
			if reason.errno != 2: # file not found
				raise
		else:
			try:
				self.__readindex()
			except Exception:
				self.__handle.close()
				raise

	def close(self):
		"""
		Closes open ChkFile class instance - no changes can be done after closing

		@return: None
		@rtype : None
		"""
		if self.__closed:
			raise ValueError('I/O operation on closed file')

		if self.__handle is not None and not self.__handle.closed:
			self.__handle.close()
			self.__handle = None

		self.__closed = True

	def __del__(self):
		"""
		Catches Instance deletion and closes filehandle

		@return: None
		@rtype : None
		"""
		if self.__handle is not None and not self.__handle.closed:
			self.__handle.close()
			self.__handle = None

		self.__closed = True

	def __readindex(self):
		"""
		Reads, parses and checks the CHK files index
		"""
		self.__handle.seek(0, 2)
		filesize = self.__handle.tell()
		self.__handle.seek(0, 0)

		try:
			num, = sread(CHK_COUNT, self.__handle)
		except struct.error:
			raise IOError('%s: short read for item count' % \
			               (self.__class__.__name__))

		if num & 0x80000000:
			if self.__passphrase is None:
				raise IOError( \
				      '%s: can\'t read encrypted file without passphrase' % \
				      (self.__class__.__name__))
		else:
			if self.__passphrase is not None:
				raise IOError( \
				      '%s: can\'t read unencrypted file with passphrase' % \
				      (self.__class__.__name__))

		num = num & self.__maxcount
		self.__handle.crypt()

		if not num:
			raise IOError('%s: empty index' % (self.__class__.__name__))

		filestart = slen(CHK_COUNT) + (num+1)*slen(CHK_INDEX)
		for cnt in xrange(0, num):
			try:
				pos, = sread(CHK_INDEX, self.__handle)
			except struct.error:
				raise IOError( \
				      '%s: short read for item pos #%d' % \
				      (self.__class__.__name__, len(self.__positions)))
			else:
				if filestart <= pos <= filesize:
					self.__positions.append(pos)
					self.__elements.append(None)
				else:
					raise IOError( \
					      '%s: item pos out of bounds (0x%08x <= 0x%08x <= 0x%08x)' % \
					      (self.__class__.__name__, filestart, pos, filesize))

		if self.__positions[0] != filestart:
			raise IOError( \
			      '%s: invalid first item pos (0x%08x != 0x%08x)' % \
			      (self.__class__.__name__, self.__positions[0], filestart))

		try:
			pos, = sread(CHK_INDEX, self.__handle)
		except struct.error:
			raise IOError( \
			      '%s: short read for end marker' % (self.__class__.__name__))
		else:
			if pos != filesize:
				raise IOError( \
				      '%s: invalid endmarker (0x%08x != 0x%08x)' % \
				      (self.__class__.__name__, pos, filesize))

		self.__handle.plain()

		for cnt, pos in enumerate(self.__positions):
			self.__handle.seek(pos, 0)
			try:
				typ, comp, dummy = sread(ITEM_HEAD, self.__handle)
			except struct.error:
				raise IOError( \
				      '%s: short read for item header #%d at 0x%08x' % \
				      (self.__class__.__name__, cnt+1, pos))
			else:
				if typ not in self.__itemtags:
					raise IOError( \
					      '%s: invalid item type #%d at 0x%08x: %02x' % \
					      (self.__class__.__name__, cnt+1, pos, typ))
				elif not self.__compressions.has_key(comp):
					raise IOError( \
					      '%s: unknown compression for item #%d at 0x%08x: %02x' % \
					      (self.__class__.__name__, cnt+1, pos, comp))
				elif cnt+1 == num:
					self.__positions[cnt] = (pos, filesize)
				else:
					self.__positions[cnt] = (pos, self.__positions[cnt+1])

	def __pack(self, data, tags):
		"""
		Packs data & tagset into StringIO buffer cache

		@param data: item data
		@type  data: String or File

		@param tags: item tags
		@type  tags: Tuple ( Int )

		@return: packed data
		@rtype : StringIO
		"""
		if len(tags) > len(self.__filetags):
			raise ValueError( \
			      '%s: too many tags (%d > %d)' % \
			      (self.__class__.__name__, len(tags), len(self.__filetags)))
		else:
			for cnt, check in enumerate(self.__filetags):
				try:
					value = tags[cnt]
				except IndexError:
					raise ValueError( \
								'%s: missing tag%d' % (self.__class__.__name__, cnt+1))
				else:
					if not (0 <= value <= check):
						raise ValueError( \
									'%s: invalid tag%d (0 <= %d <= %d)' % \
									(self.__class__.__name__, cnt+1, value, check))

			res = StringIO()
			swrite(FILE_HEAD, res, *tags)

			dtype = type(data)
			if dtype is types.StringType:
				res.write(data)
			elif dtype is types.FileType or dtype is InputType:
				pos = data.tell()
				data.seek(0, 0)
				res.write(data.read())
				data.seek(0, pos)
			else:
				raise ValueError( \
							'%s: invalid data type "%s"' % (self.__class__.__name__, dtype))

			return res

	def __read(self, index, stop=None):
		"""
		Reads data & tagset from file into StringIO buffer cache

		@param index: item index
		@type  index: Int
		@param stop: stop uncompressing when stop bytes are uncompressed
		@type  stop: Int

		@return: packed data
		@rtype : StringIO
		"""
		if self.__closed:
			raise ValueError('I/O operation on closed file')

		rstart, rend = self.__positions[index]
		self.__handle.seek(rstart, 0)

		itemtag, compression, blocks = sread(ITEM_HEAD, self.__handle)

		if itemtag not in self.__itemtags:
			raise IOError('invalid item tag found')

		if stop is None:
			stop = blocks*4
		else:
			stop = min(stop, blocks*4)

		data = StringIO()
		self.__handle.crypt()
		self.__compressions[compression][0](self.__handle, data, stop, rend)
		self.__handle.plain()
		data.seek(stop, 0)
		data.truncate()
		data.seek(0, 0)

		return data

	def save(self, force_compression=None):
		"""
		Saves cached data & tags including index to file

		@return: None
		@rtype : None
		"""
		if self.__closed:
			raise ValueError('I/O operation on closed file')

		if force_compression is not None:
			if not force_compression in self.__compressions:
				raise ValueError('Invalid compression type 0x%02x requested.' % \
				  ( force_compression ))
			else:
				compressions = [force_compression]
		else:
			compressions = self.__comp_order

		if self.__handle is None:
			outname = self.__filename
		else:
			outname = '%s.chktmp' % (self.__filename)

		handle = CipherFile(file(outname, 'wb+'), self.__passphrase)

		try:
			number = len(self)
			index  = []

			if self.__passphrase is None:
				swrite(CHK_COUNT, handle, number)
			else:
				swrite(CHK_COUNT, handle, number | 0x80000000)

			handle.write('\0'*(slen(CHK_INDEX)*(number+1)))

			for cnt in xrange(0, number):

				index.append(handle.tell())

				if self.__positions[cnt] is None:

					data = self.__elements[cnt]
					data.seek(0, 2)
					dlen = data.tell()

					outd = []
					for ctype in compressions:
						data.seek(0, 0)
						comp = StringIO()
						clen = self.__compressions[ctype][1](data, comp, dlen)
						outd.append((clen, ctype, comp))

					outd.sort()
					clen, ctyp, comp = outd.pop(0)
					del outd

					swrite(ITEM_HEAD, handle, 0x01, ctyp, (dlen+3-(dlen-1)%4)/4)
					comp.seek(0, 0)
					handle.crypt()
					handle.write(comp.read())
					handle.plain()

				else:

					pos, end = self.__positions[cnt]
					self.__handle.seek(pos, 0)
					handle.write(self.__handle.read(end-pos))

			index.append(handle.tell())

			handle.seek(slen(CHK_COUNT), 0)

			handle.crypt()
			for cnt, pos in enumerate(index):
				if pos > self.__maxpos:
					raise IOError('item position #%d out of bounds' % (cnt))
				swrite(CHK_INDEX, handle, pos)
				if cnt < number:
					index[cnt] = (pos, index[cnt+1])
			handle.plain()

			index.pop()

		except Exception:
			handle.close()
			if self.__handle is None:
				os.unlink(self.__filename)
			else:
				os.unlink(outname)
			raise

		else:
			if self.__handle is None:
				self.__handle    = handle
				self.__positions = index
				self.__changed   = False
			else:
				handle.close()
				self.__handle.close()
				os.unlink(self.__filename)
				os.rename(outname, self.__filename)
				self.__handle    = CipherFile( \
				                     file(self.__filename, 'rb+'), \
				                     self.__passphrase \
				                   )
				self.__positions = index
				self.__changed   = False

	write = save

	def count(self):
		"""
		Returns number of items in ChkFile instance

		@return: number of items
		@rtype : Int
		"""
		return len(self.__elements)

	__len__ = count

	def items(self):
		"""
		Iterator over the ChkFile instances items

		@yields: tags & data tuple
		@ytype : Tuple ( Tuple ( Int ), StringIO )
		"""
		cnt = 0
		while cnt < len(self):
			yield self.item(cnt)
			cnt += 1

	def item(self, index):
		"""
		Gets item data & tags for one item

		@param index: index number of requested item
		@type  index: Int

		@yields: tags & data tuple
		@ytype : Tuple ( Tuple ( Int ), StringIO )
		"""
		if index >= len(self):
			raise IndexError('item index out of range')

		data = self.__elements[index]
		if data is None:
			self.__elements[index] = data = self.__read(index)

		data.seek(0, 0)
		return (sread(FILE_HEAD, data), data.read())

	__getitem__ = item

	def datas(self):
		"""
		Iterator over the ChkFile instances item data

		@yields: data
		@ytype : StringIO
		"""
		cnt = 0
		while cnt < len(self):
			yield self.data(cnt)
			cnt += 1

	def data(self, index):
		"""
		Gets item data for one item

		@param index: index number of requested item
		@type  index: Int

		@yields: data
		@ytype : StringIO
		"""
		return self.item(index)[1]

	def tags(self):
		"""
		Iterator over the ChkFile instances item tags

		@yields: tags
		@ytype : Tuple ( Int )
		"""
		cnt = 0
		while cnt < len(self):
			yield self.tag(cnt)
			cnt += 1

	def tag(self, index):
		"""
		Gets item tags for one item

		@param index: index number of requested item
		@type  index: Int

		@yields: tags
		@ytype : Tuple ( Int )
		"""
		if index >= len(self):
			raise IndexError('item index out of range')

		data = self.__elements[index]
		if data is None:
			return sread(FILE_HEAD, self.__read(index, slen(FILE_HEAD)))
		else:
			data.seek(0, 0)
			return sread(FILE_HEAD, data)

	def clone(self, filename, passphrase=None):
		"""
		Clones ChkFile instance (ie. deepcopy)

		@param filename: filename of new CHK file
		@type  filename: String
		@param passphrase: passphrase to use
		@type  passphrase: String or None

		@return: ChkFile instance
		@rtype : ChkFile
		"""
		if os.path.isfile(filename):
			raise IOError('file "%s" already exists' % (filename))

		new = open(filename, passphrase)
		for tags, data in self:
			new.append(tags, data)
		return new

	def insert(self, index, tags, data):
		"""
		Inserts one item

		@param data: item data
		@type  data: String or File
		@param tags: item tags
		@type  tags: Tuple ( Int )

		@return: None
		@rtype : None
		"""
		if index > len(self):
			raise IndexError('insert item index out of range')

		if len(self) == self.__maxcount:
			raise IndexError('item count out of range')

		self.__changed = True
		self.__elements.insert(index, self.__pack(data, tags))
		self.__positions.insert(index, None)

	def delete(self, index):
		"""
		Deletes one item

		@param index: index of item to delete
		@type  index: Int

		@return: None
		@rtype : None
		"""
		if index >= len(self):
			raise IndexError('delete item index out of range')

		self.__changed = True
		del self.__elements[index]
		del self.__positions[index]

	def replace(self, index, tags, data):
		"""
		Replaces one item

		@param data: item data
		@type  data: String or File
		@param tags: item tags
		@type  tags: Tuple ( Int )

		@return: None
		@rtype : None
		"""
		if index >= len(self):
			raise IndexError('replace item index out of range')

		self.__changed = True
		self.__elements[index] = self.__pack(data, tags)
		self.__positions[index] = None

	def append(self, tags, data):
		"""
		Appends one item

		@param data: item data
		@type  data: String or File
		@param tags: item tags
		@type  tags: Tuple ( Int )

		@return: None
		@rtype : None
		"""
		if len(self) == self.__maxcount:
			raise IndexError('item count out of range')

		self.__changed = True
		self.__elements.append(self.__pack(data, tags))
		self.__positions.append(None)

	def clear(self):
		"""
		Clears ChkFile instance (ie. deletes all items)

		@return: None
		@rtype : None
		"""
		if len(self):
			self.__changed   = True
			self.__elements  = []
			self.__positions = []

